/*******************************************************************************************
 ** Raw2BMP: A simple RAW to BMP file convertor/joiner, preset for the TES4: Oblivion format
 **          (i.e. a 16-bit 1024x1024 files).
 **          Without arguments it will attempt to join: q1.raw, q2.raw, q4.raw and q3.raw in to
 **          a 2048x2048 16-bit colour BMP file called "raw2bmp.bmp".
 **          If used with file arguments, it must be: 
 **
 ** Example usage:   raw2bmp [-v] file1 [file2 .. fileN]
 **
 ** The number of files specified must be a square (e.g. 1 file, 4 files, 9 files) to form
 ** a perfect quad. This program could be modified if you want stranger combinations 
 ** (e.g. 2x3) ...
 **
 ** -v (Vality7 mode) produces a greyscale image based on the range of data available.
 **
 ** Paul Halliday: 31-Dec-2006 [ GNU License: Copy, modify, distribute as you please.;) ]
 ***************************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <math.h>

#define OUTPUT_BMP "raw2bmp.bmp"

main(int argc, char *argv[])
{
	int i,
	    p,
	    x,
	    y,
	    f = 0,
	    xfiles = 0, 
	    yfiles = 0;

	float sqrt_nfiles = 0;

	int vality_mode = 0;

	int sx = 1024,	  // Width  of expected input RAW files
	    sy = 1024,    // Height of expected input RAW files.
	    sx_out = 0,   // Final width of output BMP image.
	    sy_out = 0,	  // Final height of output BMP image.
	    bpp_in = 16,  // Bits per Pixel expected in the RAW file.
	    bpp_out = 32, // Bits per Pixel to output in the BMP file.
	    bytes_in = 2; // Bytes per Pixel expected in the RAW file (i.e. bpp_in / 8).

	unsigned short int pixel,
	    max_pixel = 0,
	    min_pixel = 65535;

	char  grey_pixel = 0;
	float scale_factor;

	int  filecount_in = 0;

	char *filename_in[64],
	     s[2048];
	
	char bmp_head[54] = {
			0x42, 0x4D, 0x98, 0xEA, 0x00, 0x00, 0x00, 0x00, 
			0x00, 0x00, 0x36, 0x00, 0x00, 0x00, 0x28, 0x00, 
			0x00, 0x00, 0xC8, 0x00, 0x00, 0x00, 0x64, 0x00, 
			0x00, 0x00, 0x01, 0x00, 0x20, 0x00, 0x00, 0x00, 
			0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x12, 0x0B, 
			0x00, 0x00, 0x12, 0x0B, 0x00, 0x00, 0x00, 0x00, 
			0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	};

	FILE *fp_in[64],
	     *fp_out;

	printf("*******************************************************************************\n"
	       "RAW2BMP: A program to convert+join TES4 RAW heightmap images into a BMP image.\n"
	       "                           by Paul Halliday (aka Lightwave)\n"
	       "*******************************************************************************\n\n");

	if (argc > 1 && strcmp(argv[1], "-v") == 0) {
		vality_mode = 1;
		printf("Running in Vality7 mode: A Greyscale image will be created.\n\n");
	}

	if (argc < (vality_mode+2)) {
		fprintf(stderr, "Since you specified no files, I will try to process the usual 4 TES4 files:\n"
			            "                Q1.raw, Q2.raw, Q4.raw and Q3.raw.\n\n");
		argc = vality_mode+5;
		argv[vality_mode+1] = "q1.raw";
		argv[vality_mode+2] = "q2.raw";
		argv[vality_mode+3] = "q4.raw";
		argv[vality_mode+4] = "q3.raw";
	} 

	while (argc-- > (vality_mode+1)) {
		filename_in[filecount_in] = argv[vality_mode+1+filecount_in];
		filecount_in++;
	}

	sqrt_nfiles = sqrt(filecount_in);
	xfiles = yfiles = sqrt_nfiles;
	if (sqrt_nfiles - (int) sqrt_nfiles > 0) {
		fprintf(stderr, "Sorry, but you need a square number of files to form a square group. e.g. 1, 4, 9\n");
		exit(1);
	}

	if ((fp_out = fopen(OUTPUT_BMP, "wb")) == 0) {
		fprintf(stderr, "Unable to open %s for writing: %s\n",
			OUTPUT_BMP, strerror(errno));
		exit(1);
	} 
	
	sx_out = sx * xfiles;  // Width  of final image.
	sy_out = sy * yfiles;  // Height of final image.
	bytes_in = bpp_in / 8;

	if (vality_mode) {
		bpp_out = 32; // A greyscale of 32-bit RGBn values if in Vality mode. 
	} else { 
		bpp_out = 16; // A straight 16-bit output of the RAW data (i.e. 16-bit colour).
	}

	memcpy(bmp_head+18, &sx_out, 4);   // Correct the header for width.
	memcpy(bmp_head+22, &sy_out, 4);   // Correct the header for height.
	memcpy(bmp_head+28, &bpp_out, 1);  // Correct the header for bits per pixel.

	// Write out the BMP header.

	for (i = 0; i < 54; i++) {
		fputc(bmp_head[i], fp_out);
	}

	// Open each input file for reading, (we've hardcoded a 64 file limit above).

	for (f = 0; f < filecount_in; f++) {
		if ((fp_in[f] = fopen(filename_in[f], "rb")) == NULL) {
			fprintf(stderr, "Unable to open %s (file number %d) for reading: %s\n",
				filename_in[f], f, strerror(errno));
			exit(1);
		} 
	}

	// In Vality mode we have to record the maximum and minimum pixel points from the RAW file
	// and calculate a scaling factor so all points lay between 0 and 255 (for greyscale).

	if (vality_mode) {
		for (y = yfiles-1; y >= 0; y--) {
			for (i = 1023; i >= 0; i--) {
				for (x = 0; x < xfiles; x++) {
					fseek(fp_in[x+(y*xfiles)], sx*bytes_in*i, SEEK_SET);
					for (p = 0; p < sx; p++) {
						fread(&pixel, bytes_in, 1, fp_in[x+(y*xfiles)]);
						if (pixel > max_pixel) max_pixel = pixel;
						if (pixel < min_pixel) min_pixel = pixel;
					}
				}
			}
		}

		scale_factor = ((float) (max_pixel - min_pixel)/(float) 255.01);

		if (scale_factor < 1) scale_factor = 1;

		printf("Highest point: %d\nLowest point:  %d\nUsing a scaling down factor of: %f\n\n", 
			max_pixel, min_pixel, scale_factor);
	}

	// A loop to cycle through all the files, copying one row at a time (sx*bytes_in = 2048 bytes = 1024 pixels of 16-bit data).

	for (y = yfiles-1; y >= 0; y--) {
		for (i = 1023; i >= 0; i--) {
			for (x = 0; x < xfiles; x++) {
				fseek(fp_in[x+(y*xfiles)], sx*bytes_in*i, SEEK_SET);
				for (p = 0; p < sx; p++) {
					fread(&pixel, bytes_in, 1, fp_in[x+(y*xfiles)]);
					if (vality_mode) {
						pixel = (unsigned char) (((float) pixel - (float) min_pixel)/ scale_factor);
						fprintf(fp_out, "%c%c%c%c", pixel, pixel, pixel, 0);
					} else {
						fwrite(&pixel, bytes_in, 1, fp_out);
					}
				}
			}
		}
	}

	// Be nice and close everything.

	for (f = 0; f < filecount_in; f++) {
		fclose(fp_in[f]);
	}
	fclose(fp_out);
	
	printf("Finished! The converted %dx%d file is called %s.\n", sx, sy, OUTPUT_BMP);

	return 0;
}
